/*
 *  trimcol - trim tab delimited columns to length
 *            and expand TAB characters according column length
 *
 *  AUTHOR: Thomas Spahni (thomas@spahni.ch)
 *  Copyright (c) 2001 by Thomas Spahni
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


/* What it does?
 * ============
 * trimcol acts as a filter. It takes a TAB delimited table
 * data from stdin and determines maximum width for every column.
 * Non printable characters are discarded.
 *
 * Note: the first line ususally contains the column names. It will
 *       be discarded but is used for counting how many columns
 *       there are.
 *
 * Output is one of two formats:
 * ============================
 * Mode 1: A list of max column-widths.
 *         Mode 1 is enabled when trimcol is used without any argument
 *
 * or:
 *
 * Mode 2: Data from stdin is output to stdout with TABs removed and 
 *         columns properly aligned. Amount of input data for mode 2
 *         is limited by buffer size (see trimcol.h). Text columns
 *         are left aligned, columns containing [0-9.] are right
 *         aligned.
 *
 * Command line options:
 * ====================
 * Note: Using command line options enables mode 2
 *
 * Arg1   A number of blanks between columns may be given;
 *        This enables mode 2 (output of aligned table).
 *
 * Arg2   <optional>. The maximum width for a column may be given.
 *        Any column text longer than this will be truncated.
 *
 * Example:
 * cat mydata.tab | trimcol 2 35 >mydata.txt
 */

#include "trimcol.h"

/* we can keep track of maximum TCNCOLS columns */
static int colwctr[TCNCOLS];

/* a flag for left align (=0) or right align (>0) */
static unsigned char colalig[TCNCOLS];


unsigned char isnotnumber(unsigned char c)
{
	switch (c) {
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
		case '.':
			return 0;
			break;
		default:
			return 1;
			break;
	}
}


int
main (int argc, char *argv[])
{
	const int newline = '\n';
	const int tabchar = '\t';

	int i;
	int c;
	int position, length, column, ncol, blanks, line;
	unsigned char isnumb;
	int maxcol;
	int intercolskip, skip;
	
	/* we keep right aligned column data in a buffer */
	unsigned char raligbuf[RALIGBUFFER];
	int rab=0, sab=0;

	size_t *bufp = 0;
	size_t *bufstart = 0, *bufend = 0;

	if (argc >= 2) {
		intercolskip = atoi(argv[1]);
		if ( intercolskip < 0 ) {
			fprintf(stderr, "trimcol: Bad argument %d\n", 
				intercolskip);
			exit(EXIT_FAILURE);
		}
		
	} else {
		intercolskip = -1;
		bufp = 0; /* avoid to test it uninitialized */
	}
                
	/* maximum length a column may have; rest is cut-off */
	if (argc >= 3) {
		maxcol = atoi(argv[2]);
	} else {
		maxcol = TCBUFFER;
	}
	
	if ( intercolskip > -1 ) {
		if ( (bufp = malloc(TCBUFFER)) == NULL ) {
			fprintf(stderr, 
			"trimcol: Could not allocate %d bytes of buffer space\n",
			TCBUFFER);
			exit(EXIT_FAILURE);
		}
		bufstart = bufp;
	}

		
	/* Array of column-width counters is reset to 0 */
	for (i=0; i < TCNCOLS; i++) {
		colwctr[i] = 0;
		colalig[i] = 0;
	}

	/*
	 * maximum width of every column is determined
	 */

	position = 0;
	length   = 0;
	column   = 0;
	ncol     = 0;  /* how many columns in the data? */

	/* Discard the first row */
	while ((c = getchar()) != newline) {
		if (length++ == TCBUFFER) {
			fprintf(stderr, "trimcol: First line overflow (%d bytes read)\n",
				TCBUFFER);
			exit(EXIT_FAILURE);
		}
		
		/* premature end of file: exit */
		if ( c == EOF ) {
			exit(EXIT_FAILURE);
		}

		/* count TABs in first row */
		if ( c == tabchar ) {
			ncol++;
		}
	}

	line   = 2;
	length = 0;
	isnumb = 0;
	while ((c = getchar()) != EOF) {
		/* give up for an unreasonable line length */
		if (length++ == TCBUFFER) {
			fprintf(stderr, "trimcol: Line overflow (%d bytes read on line %d)\n",
				TCBUFFER, line);
			exit(EXIT_FAILURE);
		}
		if ( intercolskip > -1 ) {
			/* keep a copy of stdin input */
			if ((c==tabchar) || (c==newline) || (c >= ' ')) {
				*bufp++ = (unsigned char) c;
			}
			if ( bufp >= (bufstart + TCBUFFER) ) {
				fprintf(stderr,
					"trimcol: Buffer overflow (size is %d bytes)\n",
					TCBUFFER);
				exit(EXIT_FAILURE);
			}
		}

		if (c == tabchar) {
			if ( position > colwctr[column] ) {
				colwctr[column] = position;
			}
			if (isnumb != 0) {
				colalig[column] = isnumb;
			}
			column++;
			position = 0;
			isnumb   = 0;
		} else {
			if (c == newline) {
				if ( position > colwctr[column] ) {
					colwctr[column] = position;
				}
				if (isnumb != 0) {
					colalig[column] = isnumb;
				}
				if ( column != ncol ) {
					fprintf(stderr,
					"trimcol: Data corrupt at line %d\n", line);
					exit(EXIT_FAILURE);
				}
				position = 0;
				column   = 0;
				isnumb   = 0;
				line++;
			} else {
				/* count character */
				if (c >= 32) {
					position++;
				}
				isnumb=(isnumb | isnotnumber(c));
			}
		}
	}
	if ( intercolskip > -1 ) {
		*bufp++ = EOF;
		bufend  = bufp;
	}

	/*
	 * One of the two possible output modes is selected
	 * according to the value of intercolskip
	 */

	skip = 0;
	if ( intercolskip == -1 ) /* Output Mode 1 */ {
		/* output list of column widths */
		for ( i=0; i<ncol; i++) {
			if ( i == (ncol-1) ) {
				printf("%d", colwctr[i]);
			} else {
				printf("%d ", colwctr[i]);
			}
		}
		printf("\n");
	} else { /* Output Mode 2: formatted data */
		bufp     = bufstart;
		column 	 = 0;
		position = 0;
		while ( (c = *bufp++) != EOF ) {
		    if ( colalig[column] ) {
		    /* normal left align for this column */
			if ( c == tabchar ) {
				if ( maxcol >= colwctr[column] ) {
					blanks  = colwctr[column] 
						+ intercolskip
						- position;
				} else {
					if ( position <= maxcol ) {
						blanks  = maxcol 
							+ intercolskip
							- position;
					} else {
						blanks = intercolskip;
					}
				}
				for (i = 0; i < blanks; i++) {
					putchar(' ');
				}
				position = 0;
				column++;
			} else {
				if ( c == newline ) {
					putchar(c);
					column   = 0;
					position = 0;
				} else {
					/* chars are dropped after maxcol */
					if ( position < maxcol ) {
						putchar(c);
					}
					position++;
				}
			}
		    } else {
			/* right align of numbers */
			if ( (c == tabchar) || (c == newline) ) {
				/* give out blanks, then data */
				if ( maxcol >= colwctr[column] ) {
					blanks  = colwctr[column] 
						- position;
					sab = 0;
				} else {
					if ( position <= maxcol ) {
						blanks  = maxcol 
							- position;
						sab = 0;
					} else {
						blanks = 0;
						sab = position - maxcol;
					}
				}
				for (i = 0; i < blanks; i++) {
					putchar(' ');
				}
				for (i = sab; i < rab; i++) {
					putchar(raligbuf[i]);
				}
				rab = 0;
				if ( c == newline ) {
					putchar(c);
					column   = 0;
					position = 0;
				} else {
					for (i = 0; i < intercolskip; i++) {
						putchar(' ');
					}
					position = 0;
					column++;
				}
			} else {
				raligbuf[rab++] = c;
				position++;
			}
		    }
		}

	}

	fflush(stdout);
	if ( intercolskip > -1 ) {
		free(bufstart);
	}
	exit(EXIT_SUCCESS);
}
